from uvtrick import Env 
from sklearn.datasets import make_regression

X, y = make_regression(n_samples=10_000, n_features=10, random_state=42)

def bench(X, y):
    from time import time
    from sklearn.decomposition import PCA

    tic = time()
    pca = PCA(n_components=2).fit(X)
    toc = time()
    return toc - tic

for version in ["1.4", "1.5"]:
    for i in range(4):
        timed = Env(f"scikit-learn=={version}").run(bench, X, y)
        print(version, timed)



import random
from uvtrick import Env 


def bench():
    from sklearn.datasets import make_regression
    from time import time
    from sklearn.linear_model import Ridge

    X, y = make_regression(n_samples=100_000, n_features=20, random_state=42)
    tic = time()
    pca = Ridge().fit(X, y)
    toc = time()
    return toc - tic

# for version in ["0.22", "0.23", "0.24"]:
#     for i in range(4):
#         timed = Env(f"scikit-learn=={version}", python="3.7").run(bench)
#         print(version, timed)

# for version in ["1.4", "1.5"]:
#     for i in range(4):
#         timed = Env(f"scikit-learn=={version}", python="3.9").run(bench)
#         print(version, timed)

def combiner(pairs, times=4):
    combos = []
    for pkg, py in pairs:
        for i in range(times):
            combos.append([pkg, py, i])
    random.shuffle(combos)
    return combos

combos = combiner(
    pairs=[("1.4", "3.9"), ("1.5", "3.9")], 
    times=10
)

results = []
for version, py, i in combos:
    timed = Env(f"scikit-learn=={version}", python=py).run(bench)
    results.append({
        "i": i, 
        "version": version,
        "python": py,
        "time": timed
    })
    print(version, timed)

import polars as pl

pl.DataFrame(results).sort("version").write_csv("results.csv")


from pathlib import Path

test_assets_dir = Path(__file__).parent

def test_demo():
    import demo

def test_add():
    from uvtrick import load

    script = test_assets_dir / "rich-script.py"
    add = load(script, "add")

    print(f"{add(1, 2)=}")
    print(f"{add(a=1, b=2)=}")


def test_hello():
    from uvtrick import load

    script = test_assets_dir / "rich-script.py"
    hello = load(script, "hello")
    print(f"{hello()=}")


def test_rich_versions():
    from uvtrick import Env

    rich_versions = (10, 11, 12, 13)
    print(f"Now iterating through {rich_versions=}:")


    def uses_rich(a, b):
        from importlib import metadata

        from rich import print

        version = metadata.version("rich")

        print(f"hello from rich=={version}")
        return a + b


    # This runs the function `uses_rich` in a new environment with the `rich` package installed.
    # Just like the `load` function, the result is returned via pickle.
    for version in rich_versions:
        result = Env(f"rich=={version}", python="3.12").run(uses_rich, a=1, b=2)
        print(f" --> add(1, 2) = {result}")


def test_simple():
    from uvtrick import load

    # Load the function `add` from the file `some_script.py`
    # It runs in another virtualenv, but you get back the response via pickle.
    # Be aware of the limitations, please only consider base Python objects.
    script = test_assets_dir / "some_script.py"
    add = load(script, "add")

    # This result is from the `some_script.py` file, running in another virtualenv
    # with `uv`. A pickle in a temporary file is used to communicate the result.
    result = add(1, 2)  # 3

    assert result == 3
    print(result)



from uvtrick import Env, load

hello = load("tests/rich-script.py", "hello")

print(hello())

add = load("tests/rich-script.py", "add")

print(add(1, 2))
print(add(a=1, b=2))

def uses_rich(a, b):
    from rich import print
    from importlib import metadata

    version = metadata.version("rich")
    print(f"hello from rich=={version}")
    return a + b

# This runs the function `uses_rich` in a new environment with the `rich` package installed.
# Just like the `load` function, the result is returned via pickle. 
for version in (10, 11, 12, 13):
    Env(f"rich=={version}", python="3.12").run(uses_rich, a=1, b=2)



"""When I hacks this bad, I write tests."""

import pytest

from uvtrick import Env, load

hello = load("tests/rich-script.py", "hello")
add = load("tests/rich-script.py", "add")


def test_smoke():
    assert hello() == 1


def test_args():
    assert add(1, 2) == 3
    assert add(a=1, b=4) == 5


def test_no_exist():
    with pytest.raises(ValueError):
        func = load("tests/rich-script.py", "no_exist")
        func()


def test_no_metadata():
    with pytest.raises(ValueError):
        func = load("tests/rich-fail.py", "add")
        func()


def test_env_works1():
    def uses_rich(a, b):
        from rich import print

        print("hello")
        return a + b

    for version in ["13", "12"]:
        assert Env(f"rich=={version}").run(uses_rich, a=1, b=2) == 3


def test_env_works2():
    def handles_all_types(arr, dictionary, string):
        return {"arr": arr, "dictionary": dictionary, "string": string}

    for version in ["13", "12"]:
        out = Env(f"rich=={version}").run(
            handles_all_types,
            arr=[1, 2, 3],
            dictionary={"a": 1, "b": 2},
            string="hello",
        )
        assert out == {
            "arr": [1, 2, 3],
            "dictionary": {"a": 1, "b": 2},
            "string": "hello",
        }



from __future__ import annotations

import inspect
import cloudpickle
import subprocess
import tempfile
import textwrap
from collections.abc import Callable
from pathlib import Path


def argskwargs_to_callstring(func, *args, **kwargs) -> str:
    string_kwargs = ", ".join([f"{k}={v}" for k, v in kwargs.items()])
    string_args = ", ".join([f"{a}" for a in args]) + ", " if args else ""
    return f"{func.__name__}({string_args} {string_kwargs})"


def uvtrick_(path: str | Path, func: Callable, *args, **kwargs):
    """This is a *very* hacky way to run functions from Python files from another virtual environment."""
    string_kwargs = ", ".join([f"{k}={v}" for k, v in kwargs.items()])
    string_args = ", ".join([f"{a}" for a in args]) + ", " if args else ""


    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir = Path(temp_dir)
        script = temp_dir / "pytemp.py"
        output = temp_dir / "tmp.pickle"

        code = Path(path).read_text()
        idx = code.find("if __name__")
        code = code[:idx] + "\n\n"

        if func + "(" not in code:
            raise ValueError(f"Function {func} not found in the file {path}")
        if "# /// script" not in code:
            raise ValueError("Script metadata/dependencies not found in the file")

        code += f"""if __name__ == "__main__":
    import cloudpickle
    with open('tmp.pickle', 'wb') as f:
        cloudpickle.dump({func}({string_args} {string_kwargs}), f)\n"""

        script.write_text(code)
        # print(code)
        cmd = ["uv", "run", "--with=cloudpickle", "--quiet", str(script)]
        subprocess.run(cmd, cwd=temp_dir, check=True)

        return cloudpickle.loads(output.read_bytes())


def load(path: str | Path, func: Callable) -> Callable:
    """
    Load a function from a Python file, this function will be executed in a separate virtual environment using uv.
    
    Note that this approach is more of a demo, it is very hacky and it assumes that the Python script in question 
    uses inline script metadata. More information on this feature can be found here:

    - https://docs.astral.sh/uv/guides/scripts/#declaring-script-dependencies
    - https://packaging.python.org/en/latest/specifications/inline-script-metadata/#inline-script-metadata

    Usage: 

    ```python
    from uvtrick import load

    # Load the function `hello` from the file `some_script.py`
    # It runs in another virtualenv, but you get back the response via cloudpickle. 
    # Be aware of the limitations, please only consider base Python objects.
    hello = load("some_script.py", "hello")
    ```
    """
    def load_func(*args, **kwargs):
        return uvtrick_(path, func, *args, **kwargs)
    return load_func


class Env:
    """Represents a virtual environment with a specific Python version and set of dependencies."""
    def __init__(self, *requirements: str, python: str = None, debug: bool = False):
        self.requirements = requirements
        self.python = python
        self.debug = debug
        self.temp_dir: Path = None

    @property
    def inputs(self) -> Path:
        return self.temp_dir / "pickled_inputs.pickle"

    @property
    def script(self) -> Path:
        return self.temp_dir / "pytemp.py"

    @property
    def output(self) -> Path:
        return self.temp_dir / "tmp.pickle"

    @property
    def cmd(self) -> list[str]:
        quiet = [] if self.debug else ["--quiet"]
        deps = [f"--with={dep}" for dep in self.requirements]
        pyversion = [f"--python={self.python}"] if self.python else []
        return ["uv", "run", "--with=cloudpickle", *quiet, *deps, *pyversion, str(self.script)]

    def report(self, contents: str) -> None:
        """Log the temporary dir, input kw/args and intermediate script to STDOUT."""
        print(f"Running files in {self.temp_dir}\n{self.cmd}")
        args, kwargs = cloudpickle.loads(self.inputs)
        print(f"Pickled args: {args}")
        print(f"Pickled kwargs: {kwargs}")
        print(f"Contents of the script:\n\n{contents}")
        return

    def maincall(self, func: Callable) -> str:
        """A main block to deserialise a function signature then serialise a result.

        Load the args/kwargs from an 'inputs' cloudpickle, call a Python function
        with them, and store the result in an 'output' cloudpickle.
        """
        func_name = func.__name__
        inputs_path, output_path = self.inputs, self.output
        return textwrap.dedent(f"""
        if __name__ == "__main__":
            import cloudpickle
            from pathlib import Path

            args, kwargs = cloudpickle.loads(Path('{inputs_path!s}').read_bytes())
            result = {func_name}(*args, **kwargs)
            Path('{output_path!s}').write_bytes(cloudpickle.dumps(result))
        """)

    def run(self, func: Callable, *args, **kwargs):
        """Run a function in the virtual environment using uv."""
        with tempfile.TemporaryDirectory() as temp_dir:
            self.temp_dir = Path(temp_dir)
            # First pickle the inputs
            self.inputs.write_bytes(cloudpickle.dumps((args, kwargs)))
            # Now write the contents of the script
            func_source = textwrap.dedent(inspect.getsource(func))
            contents = func_source + "\n\n" + self.maincall(func)
            self.script.write_text(contents)

            if self.debug:
                self.report(contents)
            subprocess.run(self.cmd, cwd=temp_dir, check=True)
            # Lastly load the stored result of running the script
            return cloudpickle.loads(self.output.read_bytes())
